/**
 * @file 	telnetd.c
 * @author  Oleg Antonyan <oleg.b.antonyan@gmail.com>
 * @version 1.1
 *
 * @section LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details at
 * http://www.gnu.org/copyleft/gpl.html
 *
 * @section DESCRIPTION
 *
 * Telnet-like server
 */

#include "telnetd.h"
#include "cli_wrappers.h"

struct accept_thread_params {
	int32_t socket;
	int8_t *cli_outbuf;
	struct sockaddr_in client;
	xSemaphoreHandle cli_outbuf_mutex;
	const char *PS1;
};

struct telnetd_task_params {
	uint16_t port;
	const char *username;
	const char *password;
	const char *hostname;
};


/**
 * Main telnet daemon task
 * @rapam - pointer to telnetd_task_params
 */
static void telnetd(void *parameters);

uint_fast8_t telnetd_start(void)
{
	static struct telnetd_task_params tel;
	tel.port = 23u;
	tel.username = "root";
	tel.password = "1234";
	tel.hostname = "freertos";

	portBASE_TYPE res = xTaskCreate(telnetd, "telnetd", 256u, &tel, 1, NULL);
	if(res != pdPASS)
	{
		syslog_write(LOG_ERROR, "telnetd_start()", "Cannot create telnetd task");
		return 0;
	}
	cli_wrappers_init();
	return 1u;
}

#define SEND_NEWLINE(sock) send(sock, "\r\n", 2, 0)

static void accept_thread(void *parameters)
{
	struct accept_thread_params *params = (struct accept_thread_params *)parameters;

	int32_t sockfd =  params->socket;
	struct sockaddr_in client_addr = params->client;
	int8_t *cli_outbuf = params->cli_outbuf;
	xSemaphoreHandle cli_outbuf_mutex = params->cli_outbuf_mutex;
	char PS1[configMAX_TASK_NAME_LEN] = {0};
	strcpy(PS1, params->PS1);

	const uint_fast32_t conn_timeout_period = 300000u;
	setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, &conn_timeout_period, sizeof conn_timeout_period);

	int32_t bytes_recv = 0;
	char cli_input[64] = {0};
	int8_t in_char = 0;

	syslog_write(LOG_INFORMATION, pcTaskGetTaskName(NULL), "Connection accepted");

	SEND_NEWLINE(sockfd);

	if(get_string_uptime(cli_input, sizeof cli_input))
	{
		send(sockfd, cli_input, sizeof cli_input, 0);

		SEND_NEWLINE(sockfd);
		SEND_NEWLINE(sockfd);
	}

	uint8_t cli_index = 0;
	memset(cli_input, 0, sizeof cli_input);

	send(sockfd, PS1, strlen(PS1), 0);
	do
	{
		bytes_recv = recv(sockfd, &in_char, sizeof(in_char), 0);
		if(bytes_recv > 0)
		{
			if(in_char == '\n')
			{
				if(strcmp("quit", cli_input) == 0)
				{
					bytes_recv = 0;
				}
				else
				{
					/* The input string was not a quit command.
					 * Pass the string to the command interpreter.
					 * Transmit a line separator, just to make the
					 * output easier to read. */
				    SEND_NEWLINE(sockfd);
				    portBASE_TYPE res = pdFALSE;
					do
					{
						/* Ensure there is not a string lingering in
						 * the output buffer. */
						cli_input[cli_index] = '\0';
						xSemaphoreTake(cli_outbuf_mutex, portMAX_DELAY);
						res = FreeRTOS_CLIProcessCommand(cli_input, cli_outbuf, configCOMMAND_INT_MAX_OUTPUT_SIZE);
						send(sockfd, cli_outbuf, strlen(cli_outbuf), 0);
						xSemaphoreGive(cli_outbuf_mutex);

					} while(res != pdFALSE);

					/* All the strings generated by the input
					command have been sent.  Clear the input
					string ready to receive the next command. */
					cli_index = 0;
					memset(cli_input, 0, sizeof cli_input);
					SEND_NEWLINE(sockfd);
				}
				send(sockfd, PS1, strlen(PS1), 0);
			}
			else
			{
				if(in_char == '\r')
				{
					/* Ignore the character. */
				}
				else if(in_char == '\b')
				{
					/* Backspace was pressed.  Erase the last
					character in the string - if any. */
					if(cli_index > 0)
					{
						cli_index--;
						cli_input[cli_index] = '\0';
					}
				}
				else
				{
					/* A character was entered.  Add it to the string
					entered so far.  When a \n is entered the complete
					string will be passed to the command interpreter. */
					if(cli_index < sizeof cli_input)
					{
						cli_input[ cli_index ] = in_char;
						cli_index++;
					}
				}
			}
		}

	} while(bytes_recv > 0);

	syslog_write(LOG_INFORMATION, pcTaskGetTaskName(NULL), "Connection closed");
	close(sockfd);
	vTaskDelete(NULL);
}

static uint_fast8_t get_login_or_password(int32_t socket, const char * const valid, char *buf, size_t size)
{
	if(!buf || size < MIN_LOGIN_PASS_LENGTH)
	{
		return 0;
	}
	int32_t bytes = 0;
	char inchar = 0;
	uint_fast8_t index = 0;
	uint_fast8_t result = 0;
	do
	{
		bytes = recv(socket, &inchar, sizeof(inchar), 0);

		if(bytes > 0)
		{
			if(inchar == '\n')
			{
				buf[--index] = '\0';
				if(valid == NULL)
				{
					bytes = 0;
				}
				if(strcmp(buf, valid) == 0)
				{
					result = 1u;
					bytes = 0;
				}
				else
				{
					bytes = 0;
				}
			}
			else
			{
				if(inchar == '\b')
				{
					if(index > 0)
					{
						index--;
						buf[index] = '\0';
					}
				}
				else
				{
					if(index < size)
					{
						buf[index] = inchar;
						index++;
					}
					else
					{
						bytes = 0;
					}
				}
			}
		}

	} while(bytes > 0);

	return result;
}

static void telnetd(void *parameters)
{
	struct telnetd_task_params *params = (struct telnetd_task_params *)parameters;

	uint16_t port = params->port;
	const char *username = params->username;
	const char *password = params->password;
	const char *hostname = params->hostname;

	syslog_write(LOG_INFORMATION, pcTaskGetTaskName(NULL), "Up and running at port %u", port);

	int32_t sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if(sockfd < 0)
	{
		syslog_write(LOG_ERROR, pcTaskGetTaskName(NULL), "Cannot create socket. %d", sockfd);
		vTaskDelete(NULL);
	}
	struct sockaddr_in local_addr, client_addr;

	memset((void *)&local_addr, 0, sizeof(local_addr));
	local_addr.sin_family = AF_INET;
	local_addr.sin_len = sizeof(local_addr);
	local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	local_addr.sin_port = ntohs(((uint16_t)port));

	if(bind(sockfd, (struct sockaddr *)&local_addr, sizeof(local_addr)) < 0)
	{
		syslog_write(LOG_ERROR, pcTaskGetTaskName(NULL), "Cannot bind to %s:%u", inet_ntoa(local_addr.sin_addr), port);
		close(sockfd);
		vTaskDelete(NULL);
	}

	if(listen(sockfd, 20) != 0)
	{
		syslog_write(LOG_ERROR, pcTaskGetTaskName(NULL), "Cannot listen %s:%u", inet_ntoa(local_addr.sin_addr), port);
		close(sockfd);
		vTaskDelete(NULL);
	}

	socklen_t addr_len = sizeof(client_addr);

	uint_fast8_t exit_flag = 0;

	int8_t *cli_outbuf = FreeRTOS_CLIGetOutputBuffer();
	xSemaphoreHandle cli_outbuf_mutex = xSemaphoreCreateMutex();

	do
	{
		int32_t sockfd_client = 0;
		memset((void *)&client_addr, 0, sizeof(client_addr));
		sockfd_client = accept(sockfd, (struct sockaddr *)&client_addr, &addr_len);
		if(sockfd_client >= 0)
		{
			syslog_write(LOG_INFORMATION, pcTaskGetTaskName(NULL), "Client %s:%u connected", inet_ntoa(client_addr.sin_addr), client_addr.sin_port);

			const char * const welcome = "\r\nWelcome to FreeRTOS telnet server\r\nPress enter to continue\r\n";
			const char * const welcome_login = "login : ";
			const char * const welcome_password = "password : ";
			const char * const login_fail = "Login incorrect\r\n";
			const char * const login_ok = "Login successful\r\n";

			send(sockfd_client, welcome, strlen(welcome), 0);

			const uint_fast32_t login_timeout_period = 30000u;

			setsockopt(sockfd_client, SOL_SOCKET, SO_RCVTIMEO, &login_timeout_period, sizeof(login_timeout_period));

			char in = 0;
			int32_t b = 0;
			do
			{
				b = recv(sockfd_client, &in, sizeof(in), 0);
			} while(in != '\n' && b > 0);

			send(sockfd_client, welcome_login, strlen(welcome_login), 0);

			char login_user[MAX_LOGIN_PASS_LENGTH] = {0};
			char login_passwd[MAX_LOGIN_PASS_LENGTH] = {0};

			uint_fast8_t userok = get_login_or_password(sockfd_client, username, login_user, sizeof login_user);
			if(userok)
			{
				syslog_write(LOG_INFORMATION, pcTaskGetTaskName(NULL), "Login \"%s\" accepted", login_user);
			}
			else
			{
				syslog_write(LOG_WARNING, pcTaskGetTaskName(NULL), "Wrong login \"%s\"", login_user);
			}

			send(sockfd_client, welcome_password, strlen(welcome_password), 0);

			uint_fast8_t passok = get_login_or_password(sockfd_client, password, login_passwd, sizeof login_passwd);
			if(passok)
			{
				syslog_write(LOG_INFORMATION, pcTaskGetTaskName(NULL), "Password accepted");
			}
			else
			{
				syslog_write(LOG_WARNING, pcTaskGetTaskName(NULL), "Wrong password");
			}

			if(userok && passok)
			{
				syslog_write(LOG_INFORMATION, pcTaskGetTaskName(NULL), "Successful authentification \"%s\" from %s:%u", username, inet_ntoa(client_addr.sin_addr), client_addr.sin_port);

				send(sockfd_client, login_ok, strlen(login_ok), 0);

				struct accept_thread_params params;
				params.cli_outbuf = cli_outbuf;
				params.client = client_addr;
				params.socket = sockfd_client;
				params.cli_outbuf_mutex = cli_outbuf_mutex;
				char PS1[configMAX_TASK_NAME_LEN] = {0};
				sprintf(PS1, "%s@%s $ ", login_user, hostname);
				params.PS1 = PS1;
				char task_name[configMAX_TASK_NAME_LEN] = {0};
				sprintf(task_name, "%s client %s@%s:%u", pcTaskGetTaskName(NULL), login_user, inet_ntoa(client_addr.sin_addr), client_addr.sin_port);
				xTaskCreate(accept_thread, (const signed portCHAR *)task_name, 256u , (void *)&params, 1, NULL);
				taskYIELD();
			}
			else
			{
				const uint_fast32_t block = BLOCK_TIME_AFTER_FAILED_LOGIN * 1000;
				syslog_write(LOG_WARNING, pcTaskGetTaskName(NULL), "Authentification failed. Repeat after %u seconds", block / 1000);

				send(sockfd_client, login_fail, strlen(login_fail), 0);
				close(sockfd_client);
				sockfd_client = 0;
				vTaskDelay(block / portTICK_RATE_MS);
			}
		}
	}
	while(!exit_flag);

	syslog_write(LOG_WARNING, pcTaskGetTaskName(NULL), "Task closed");
	close(sockfd);
	vSemaphoreDelete(cli_outbuf_mutex);
	vTaskDelete(NULL);
}
